time: float = 0.0
pp8_mrna_degradation_rate: float = 1.00000000000000
pp9_mrna_degradation_rate: float = 1.00000000000000
p1_degradation_rate: float = 0.500000000000000
p2_degradation_rate: float = 0.500000000000000
p3_degradation_rate: float = 0.500000000000000
p4_degradation_rate: float = 0.500000000000000
p5_degradation_rate: float = 0.500000000000000
p6_degradation_rate: float = 0.500000000000000
p7_degradation_rate: float = 0.500000000000000
p8_degradation_rate: float = 0.500000000000000
p9_degradation_rate: float = 0.500000000000000
v1_Kd: float = 11.1470000000000
v1_h: float = 1.00000000000000
v2_Kd: float = 1.00000000000000
v2_h: float = 4.00000000000000
v3_Kd: float = 20.0000000000000
v3_h: float = 1.00000000000000
v4_Kd: float = 0.200000000000000
v4_h: float = 4.00000000000000
v5_Kd: float = 0.200000000000000
v5_h: float = 4.00000000000000
v6_Kd: float = 0.0400000000000000
v6_h: float = 4.00000000000000
v7_Kd: float = 0.0200000000000000
v7_h: float = 4.00000000000000
v8_Kd: float = 0.0400000000000000
v8_h: float = 4.00000000000000
v9_Kd: float = 0.200000000000000
v9_h: float = 4.00000000000000
pp1_mrna_degradation_rate: float = 1.00000000000000
pp2_mrna_degradation_rate: float = 1.00000000000000
pro1_strength: float = 2.00000000000000
pro2_strength: float = 4.50770000000000
pro3_strength: float = 5.00000000000000
pro4_strength: float = 5.00000000000000
pro5_strength: float = 5.00000000000000
pro6_strength: float = 1.31000000000000
pro7_strength: float = 1.31000000000000
pro8_strength: float = 5.00000000000000
pro9_strength: float = 5.00000000000000
pp3_mrna_degradation_rate: float = 1.00000000000000
v10_Kd: float = 0.0200000000000000
v10_h: float = 4.00000000000000
v11_Kd: float = 0.100000000000000
v11_h: float = 2.00000000000000
v12_Kd: float = 0.100000000000000
v12_h: float = 2.00000000000000
v13_Kd: float = 0.0100000000000000
v13_h: float = 2.00000000000000
pp4_mrna_degradation_rate: float = 1.00000000000000
v14_Kd: float = 1.00000000000000
v14_h: float = 4.00000000000000
v15_Kd: float = 20.0000000000000
v15_h: float = 1.00000000000000
pp5_mrna_degradation_rate: float = 1.00000000000000
rbs1_strength: float = 0.366800000000000
rbs2_strength: float = 1.41020000000000
rbs3_strength: float = 0.800000000000000
rbs4_strength: float = 2.21000000000000
rbs5_strength: float = 0.500000000000000
rbs6_strength: float = 2.00000000000000
rbs7_strength: float = 5.00000000000000
rbs8_strength: float = 3.63770000000000
rbs9_strength: float = 8.00000000000000
pp6_mrna_degradation_rate: float = 1.00000000000000
pp7_mrna_degradation_rate: float = 1.00000000000000
DefaultCompartment: float = 1.00000000000000
pp9_mrna: float = 0.0
p9: float = 0.0
pp8_mrna: float = 0.0
p8: float = 0.0
pp7_mrna: float = 0.0
p7: float = 0.0
pp6_mrna: float = 0.0
p6: float = 0.0
pp5_mrna: float = 0.0
p5: float = 0.0
pp4_mrna: float = 0.0
p4: float = 0.0
pp3_mrna: float = 0.0
p3: float = 0.0
pp2_mrna: float = 0.0
p2: float = 0.0
pp1_mrna: float = 0.0
p1: float = 5.00000000000000

# Initial assignments
as8 = (p5 / v9_Kd) ** v9_h / (DefaultCompartment * ((p5 / v9_Kd) ** v9_h + 1))
cod1 = pro1_strength
rs1 = 1.0 / (DefaultCompartment * ((p9 / v13_Kd) ** v13_h + 1))
rs2 = 1.0 / (DefaultCompartment * ((p2 / v2_Kd) ** v2_h + 1))
rs3 = 1.0 / (DefaultCompartment * ((p3 / v3_Kd) ** v3_h + 1))
rs4 = 1.0 / (DefaultCompartment * ((p8 / v11_Kd) ** v11_h + 1))
rs5 = 1.0 / (DefaultCompartment * ((p8 / v12_Kd) ** v12_h + 1))
rs6 = 1.0 / (DefaultCompartment * ((p2 / v14_Kd) ** v14_h + 1))
rs7 = 1.0 / (DefaultCompartment * ((p3 / v15_Kd) ** v15_h + 1))
as1 = (p1 / v1_Kd) ** v1_h / (DefaultCompartment * ((p1 / v1_Kd) ** v1_h + 1))
as2 = (p4 / v4_Kd) ** v4_h / (DefaultCompartment * ((p4 / v4_Kd) ** v4_h + 1))
as3 = (p5 / v5_Kd) ** v5_h / (DefaultCompartment * ((p5 / v5_Kd) ** v5_h + 1))
as4 = (p6 / v6_Kd) ** v6_h / (DefaultCompartment * ((p6 / v6_Kd) ** v6_h + 1))
as5 = (p7 / v7_Kd) ** v7_h / (DefaultCompartment * ((p7 / v7_Kd) ** v7_h + 1))
as6 = (p7 / v10_Kd) ** v10_h / (DefaultCompartment * ((p7 / v10_Kd) ** v10_h + 1))
as7 = (p6 / v8_Kd) ** v8_h / (DefaultCompartment * ((p6 / v8_Kd) ** v8_h + 1))
pp9_mrna_conc = pp9_mrna / DefaultCompartment
p9_conc = p9 / DefaultCompartment
pp8_mrna_conc = pp8_mrna / DefaultCompartment
p8_conc = p8 / DefaultCompartment
pp7_mrna_conc = pp7_mrna / DefaultCompartment
p7_conc = p7 / DefaultCompartment
pp6_mrna_conc = pp6_mrna / DefaultCompartment
p6_conc = p6 / DefaultCompartment
pp5_mrna_conc = pp5_mrna / DefaultCompartment
p5_conc = p5 / DefaultCompartment
pp4_mrna_conc = pp4_mrna / DefaultCompartment
p4_conc = p4 / DefaultCompartment
pp3_mrna_conc = pp3_mrna / DefaultCompartment
p3_conc = p3 / DefaultCompartment
pp2_mrna_conc = pp2_mrna / DefaultCompartment
p2_conc = p2 / DefaultCompartment
pp1_mrna_conc = pp1_mrna / DefaultCompartment
p1_conc = p1 / DefaultCompartment
pp9_v2 = pp9_mrna_conc * pp9_mrna_degradation_rate
pp9_v3 = pp9_mrna_conc * rbs9_strength
pp9_v4 = p9_conc * p9_degradation_rate
pp8_v2 = pp8_mrna_conc * pp8_mrna_degradation_rate
pp8_v3 = pp8_mrna_conc * rbs8_strength
pp8_v4 = p8_conc * p8_degradation_rate
pp7_v2 = pp7_mrna_conc * pp7_mrna_degradation_rate
pp7_v3 = pp7_mrna_conc * rbs7_strength
pp7_v4 = p7_conc * p7_degradation_rate
pp6_v2 = pp6_mrna_conc * pp6_mrna_degradation_rate
pp6_v3 = pp6_mrna_conc * rbs6_strength
pp6_v4 = p6_conc * p6_degradation_rate
pp5_v2 = pp5_mrna_conc * pp5_mrna_degradation_rate
pp5_v3 = pp5_mrna_conc * rbs5_strength
pp5_v4 = p5_conc * p5_degradation_rate
pp4_v2 = pp4_mrna_conc * pp4_mrna_degradation_rate
pp4_v3 = pp4_mrna_conc * rbs4_strength
pp4_v4 = p4_conc * p4_degradation_rate
pp3_v2 = pp3_mrna_conc * pp3_mrna_degradation_rate
pp3_v3 = pp3_mrna_conc * rbs3_strength
pp3_v4 = p3_conc * p3_degradation_rate
pp2_v2 = pp2_mrna_conc * pp2_mrna_degradation_rate
pp2_v3 = pp2_mrna_conc * rbs2_strength
pp2_v4 = p2_conc * p2_degradation_rate
pp1_v1 = cod1
pp1_v2 = pp1_mrna_conc * pp1_mrna_degradation_rate
pp1_v3 = pp1_mrna_conc * rbs1_strength
pp1_v4 = p1_conc * p1_degradation_rate
cod2 = as1 * pro2_strength * rs1
cod3 = pro3_strength * rs2 * rs3
cod4 = pro4_strength * rs6 * rs7
cod5 = as2 * pro5_strength
cod6 = pro6_strength * (as3 + as4)
cod7 = pro7_strength * (as7 + as8)
cod8 = as5 * pro8_strength * rs4
cod9 = as6 * pro9_strength * rs5
pp9_v1 = cod9
pp8_v1 = cod8
pp7_v1 = cod7
pp6_v1 = cod6
pp5_v1 = cod5
pp4_v1 = cod4
pp3_v1 = cod3
pp2_v1 = cod2
y0 = [
    p1,
    p2,
    p3,
    p4,
    p5,
    p6,
    p7,
    p8,
    p9,
    pp1_mrna,
    pp2_mrna,
    pp3_mrna,
    pp4_mrna,
    pp5_mrna,
    pp6_mrna,
    pp7_mrna,
    pp8_mrna,
    pp9_mrna,
]
variable_names = [
    "p1",
    "p2",
    "p3",
    "p4",
    "p5",
    "p6",
    "p7",
    "p8",
    "p9",
    "pp1_mrna",
    "pp2_mrna",
    "pp3_mrna",
    "pp4_mrna",
    "pp5_mrna",
    "pp6_mrna",
    "pp7_mrna",
    "pp8_mrna",
    "pp9_mrna",
]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (
        p1,
        p2,
        p3,
        p4,
        p5,
        p6,
        p7,
        p8,
        p9,
        pp1_mrna,
        pp2_mrna,
        pp3_mrna,
        pp4_mrna,
        pp5_mrna,
        pp6_mrna,
        pp7_mrna,
        pp8_mrna,
        pp9_mrna,
    ) = variables
    as8: float = (p5 / v9_Kd) ** v9_h / (
        DefaultCompartment * ((p5 / v9_Kd) ** v9_h + 1)
    )
    cod1: float = pro1_strength
    rs1: float = 1.0 / (DefaultCompartment * ((p9 / v13_Kd) ** v13_h + 1))
    rs2: float = 1.0 / (DefaultCompartment * ((p2 / v2_Kd) ** v2_h + 1))
    rs3: float = 1.0 / (DefaultCompartment * ((p3 / v3_Kd) ** v3_h + 1))
    rs4: float = 1.0 / (DefaultCompartment * ((p8 / v11_Kd) ** v11_h + 1))
    rs5: float = 1.0 / (DefaultCompartment * ((p8 / v12_Kd) ** v12_h + 1))
    rs6: float = 1.0 / (DefaultCompartment * ((p2 / v14_Kd) ** v14_h + 1))
    rs7: float = 1.0 / (DefaultCompartment * ((p3 / v15_Kd) ** v15_h + 1))
    as1: float = (p1 / v1_Kd) ** v1_h / (
        DefaultCompartment * ((p1 / v1_Kd) ** v1_h + 1)
    )
    as2: float = (p4 / v4_Kd) ** v4_h / (
        DefaultCompartment * ((p4 / v4_Kd) ** v4_h + 1)
    )
    as3: float = (p5 / v5_Kd) ** v5_h / (
        DefaultCompartment * ((p5 / v5_Kd) ** v5_h + 1)
    )
    as4: float = (p6 / v6_Kd) ** v6_h / (
        DefaultCompartment * ((p6 / v6_Kd) ** v6_h + 1)
    )
    as5: float = (p7 / v7_Kd) ** v7_h / (
        DefaultCompartment * ((p7 / v7_Kd) ** v7_h + 1)
    )
    as6: float = (p7 / v10_Kd) ** v10_h / (
        DefaultCompartment * ((p7 / v10_Kd) ** v10_h + 1)
    )
    as7: float = (p6 / v8_Kd) ** v8_h / (
        DefaultCompartment * ((p6 / v8_Kd) ** v8_h + 1)
    )
    pp9_mrna_conc: float = pp9_mrna / DefaultCompartment
    p9_conc: float = p9 / DefaultCompartment
    pp8_mrna_conc: float = pp8_mrna / DefaultCompartment
    p8_conc: float = p8 / DefaultCompartment
    pp7_mrna_conc: float = pp7_mrna / DefaultCompartment
    p7_conc: float = p7 / DefaultCompartment
    pp6_mrna_conc: float = pp6_mrna / DefaultCompartment
    p6_conc: float = p6 / DefaultCompartment
    pp5_mrna_conc: float = pp5_mrna / DefaultCompartment
    p5_conc: float = p5 / DefaultCompartment
    pp4_mrna_conc: float = pp4_mrna / DefaultCompartment
    p4_conc: float = p4 / DefaultCompartment
    pp3_mrna_conc: float = pp3_mrna / DefaultCompartment
    p3_conc: float = p3 / DefaultCompartment
    pp2_mrna_conc: float = pp2_mrna / DefaultCompartment
    p2_conc: float = p2 / DefaultCompartment
    pp1_mrna_conc: float = pp1_mrna / DefaultCompartment
    p1_conc: float = p1 / DefaultCompartment
    pp9_v2: float = pp9_mrna_conc * pp9_mrna_degradation_rate
    pp9_v3: float = pp9_mrna_conc * rbs9_strength
    pp9_v4: float = p9_conc * p9_degradation_rate
    pp8_v2: float = pp8_mrna_conc * pp8_mrna_degradation_rate
    pp8_v3: float = pp8_mrna_conc * rbs8_strength
    pp8_v4: float = p8_conc * p8_degradation_rate
    pp7_v2: float = pp7_mrna_conc * pp7_mrna_degradation_rate
    pp7_v3: float = pp7_mrna_conc * rbs7_strength
    pp7_v4: float = p7_conc * p7_degradation_rate
    pp6_v2: float = pp6_mrna_conc * pp6_mrna_degradation_rate
    pp6_v3: float = pp6_mrna_conc * rbs6_strength
    pp6_v4: float = p6_conc * p6_degradation_rate
    pp5_v2: float = pp5_mrna_conc * pp5_mrna_degradation_rate
    pp5_v3: float = pp5_mrna_conc * rbs5_strength
    pp5_v4: float = p5_conc * p5_degradation_rate
    pp4_v2: float = pp4_mrna_conc * pp4_mrna_degradation_rate
    pp4_v3: float = pp4_mrna_conc * rbs4_strength
    pp4_v4: float = p4_conc * p4_degradation_rate
    pp3_v2: float = pp3_mrna_conc * pp3_mrna_degradation_rate
    pp3_v3: float = pp3_mrna_conc * rbs3_strength
    pp3_v4: float = p3_conc * p3_degradation_rate
    pp2_v2: float = pp2_mrna_conc * pp2_mrna_degradation_rate
    pp2_v3: float = pp2_mrna_conc * rbs2_strength
    pp2_v4: float = p2_conc * p2_degradation_rate
    pp1_v1: float = cod1
    pp1_v2: float = pp1_mrna_conc * pp1_mrna_degradation_rate
    pp1_v3: float = pp1_mrna_conc * rbs1_strength
    pp1_v4: float = p1_conc * p1_degradation_rate
    cod2: float = as1 * pro2_strength * rs1
    cod3: float = pro3_strength * rs2 * rs3
    cod4: float = pro4_strength * rs6 * rs7
    cod5: float = as2 * pro5_strength
    cod6: float = pro6_strength * (as3 + as4)
    cod7: float = pro7_strength * (as7 + as8)
    cod8: float = as5 * pro8_strength * rs4
    cod9: float = as6 * pro9_strength * rs5
    pp9_v1: float = cod9
    pp8_v1: float = cod8
    pp7_v1: float = cod7
    pp6_v1: float = cod6
    pp5_v1: float = cod5
    pp4_v1: float = cod4
    pp3_v1: float = cod3
    pp2_v1: float = cod2
    dpp9_mrnadt: float = DefaultCompartment * pp9_v1 - DefaultCompartment * pp9_v2
    dp9dt: float = DefaultCompartment * pp9_v3 - DefaultCompartment * pp9_v4
    dpp8_mrnadt: float = DefaultCompartment * pp8_v1 - DefaultCompartment * pp8_v2
    dp8dt: float = DefaultCompartment * pp8_v3 - DefaultCompartment * pp8_v4
    dpp7_mrnadt: float = DefaultCompartment * pp7_v1 - DefaultCompartment * pp7_v2
    dp7dt: float = DefaultCompartment * pp7_v3 - DefaultCompartment * pp7_v4
    dpp6_mrnadt: float = DefaultCompartment * pp6_v1 - DefaultCompartment * pp6_v2
    dp6dt: float = DefaultCompartment * pp6_v3 - DefaultCompartment * pp6_v4
    dpp5_mrnadt: float = DefaultCompartment * pp5_v1 - DefaultCompartment * pp5_v2
    dp5dt: float = DefaultCompartment * pp5_v3 - DefaultCompartment * pp5_v4
    dpp4_mrnadt: float = DefaultCompartment * pp4_v1 - DefaultCompartment * pp4_v2
    dp4dt: float = DefaultCompartment * pp4_v3 - DefaultCompartment * pp4_v4
    dpp3_mrnadt: float = DefaultCompartment * pp3_v1 - DefaultCompartment * pp3_v2
    dp3dt: float = DefaultCompartment * pp3_v3 - DefaultCompartment * pp3_v4
    dpp2_mrnadt: float = DefaultCompartment * pp2_v1 - DefaultCompartment * pp2_v2
    dp2dt: float = DefaultCompartment * pp2_v3 - DefaultCompartment * pp2_v4
    dpp1_mrnadt: float = DefaultCompartment * pp1_v1 - DefaultCompartment * pp1_v2
    dp1dt: float = DefaultCompartment * pp1_v3 - DefaultCompartment * pp1_v4
    return (
        dp1dt,
        dp2dt,
        dp3dt,
        dp4dt,
        dp5dt,
        dp6dt,
        dp7dt,
        dp8dt,
        dp9dt,
        dpp1_mrnadt,
        dpp2_mrnadt,
        dpp3_mrnadt,
        dpp4_mrnadt,
        dpp5_mrnadt,
        dpp6_mrnadt,
        dpp7_mrnadt,
        dpp8_mrnadt,
        dpp9_mrnadt,
    )


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (
        p1,
        p2,
        p3,
        p4,
        p5,
        p6,
        p7,
        p8,
        p9,
        pp1_mrna,
        pp2_mrna,
        pp3_mrna,
        pp4_mrna,
        pp5_mrna,
        pp6_mrna,
        pp7_mrna,
        pp8_mrna,
        pp9_mrna,
    ) = variables
    as8: float = (p5 / v9_Kd) ** v9_h / (
        DefaultCompartment * ((p5 / v9_Kd) ** v9_h + 1)
    )
    cod1: float = pro1_strength
    rs1: float = 1.0 / (DefaultCompartment * ((p9 / v13_Kd) ** v13_h + 1))
    rs2: float = 1.0 / (DefaultCompartment * ((p2 / v2_Kd) ** v2_h + 1))
    rs3: float = 1.0 / (DefaultCompartment * ((p3 / v3_Kd) ** v3_h + 1))
    rs4: float = 1.0 / (DefaultCompartment * ((p8 / v11_Kd) ** v11_h + 1))
    rs5: float = 1.0 / (DefaultCompartment * ((p8 / v12_Kd) ** v12_h + 1))
    rs6: float = 1.0 / (DefaultCompartment * ((p2 / v14_Kd) ** v14_h + 1))
    rs7: float = 1.0 / (DefaultCompartment * ((p3 / v15_Kd) ** v15_h + 1))
    as1: float = (p1 / v1_Kd) ** v1_h / (
        DefaultCompartment * ((p1 / v1_Kd) ** v1_h + 1)
    )
    as2: float = (p4 / v4_Kd) ** v4_h / (
        DefaultCompartment * ((p4 / v4_Kd) ** v4_h + 1)
    )
    as3: float = (p5 / v5_Kd) ** v5_h / (
        DefaultCompartment * ((p5 / v5_Kd) ** v5_h + 1)
    )
    as4: float = (p6 / v6_Kd) ** v6_h / (
        DefaultCompartment * ((p6 / v6_Kd) ** v6_h + 1)
    )
    as5: float = (p7 / v7_Kd) ** v7_h / (
        DefaultCompartment * ((p7 / v7_Kd) ** v7_h + 1)
    )
    as6: float = (p7 / v10_Kd) ** v10_h / (
        DefaultCompartment * ((p7 / v10_Kd) ** v10_h + 1)
    )
    as7: float = (p6 / v8_Kd) ** v8_h / (
        DefaultCompartment * ((p6 / v8_Kd) ** v8_h + 1)
    )
    pp9_mrna_conc: float = pp9_mrna / DefaultCompartment
    p9_conc: float = p9 / DefaultCompartment
    pp8_mrna_conc: float = pp8_mrna / DefaultCompartment
    p8_conc: float = p8 / DefaultCompartment
    pp7_mrna_conc: float = pp7_mrna / DefaultCompartment
    p7_conc: float = p7 / DefaultCompartment
    pp6_mrna_conc: float = pp6_mrna / DefaultCompartment
    p6_conc: float = p6 / DefaultCompartment
    pp5_mrna_conc: float = pp5_mrna / DefaultCompartment
    p5_conc: float = p5 / DefaultCompartment
    pp4_mrna_conc: float = pp4_mrna / DefaultCompartment
    p4_conc: float = p4 / DefaultCompartment
    pp3_mrna_conc: float = pp3_mrna / DefaultCompartment
    p3_conc: float = p3 / DefaultCompartment
    pp2_mrna_conc: float = pp2_mrna / DefaultCompartment
    p2_conc: float = p2 / DefaultCompartment
    pp1_mrna_conc: float = pp1_mrna / DefaultCompartment
    p1_conc: float = p1 / DefaultCompartment
    pp9_v2: float = pp9_mrna_conc * pp9_mrna_degradation_rate
    pp9_v3: float = pp9_mrna_conc * rbs9_strength
    pp9_v4: float = p9_conc * p9_degradation_rate
    pp8_v2: float = pp8_mrna_conc * pp8_mrna_degradation_rate
    pp8_v3: float = pp8_mrna_conc * rbs8_strength
    pp8_v4: float = p8_conc * p8_degradation_rate
    pp7_v2: float = pp7_mrna_conc * pp7_mrna_degradation_rate
    pp7_v3: float = pp7_mrna_conc * rbs7_strength
    pp7_v4: float = p7_conc * p7_degradation_rate
    pp6_v2: float = pp6_mrna_conc * pp6_mrna_degradation_rate
    pp6_v3: float = pp6_mrna_conc * rbs6_strength
    pp6_v4: float = p6_conc * p6_degradation_rate
    pp5_v2: float = pp5_mrna_conc * pp5_mrna_degradation_rate
    pp5_v3: float = pp5_mrna_conc * rbs5_strength
    pp5_v4: float = p5_conc * p5_degradation_rate
    pp4_v2: float = pp4_mrna_conc * pp4_mrna_degradation_rate
    pp4_v3: float = pp4_mrna_conc * rbs4_strength
    pp4_v4: float = p4_conc * p4_degradation_rate
    pp3_v2: float = pp3_mrna_conc * pp3_mrna_degradation_rate
    pp3_v3: float = pp3_mrna_conc * rbs3_strength
    pp3_v4: float = p3_conc * p3_degradation_rate
    pp2_v2: float = pp2_mrna_conc * pp2_mrna_degradation_rate
    pp2_v3: float = pp2_mrna_conc * rbs2_strength
    pp2_v4: float = p2_conc * p2_degradation_rate
    pp1_v1: float = cod1
    pp1_v2: float = pp1_mrna_conc * pp1_mrna_degradation_rate
    pp1_v3: float = pp1_mrna_conc * rbs1_strength
    pp1_v4: float = p1_conc * p1_degradation_rate
    cod2: float = as1 * pro2_strength * rs1
    cod3: float = pro3_strength * rs2 * rs3
    cod4: float = pro4_strength * rs6 * rs7
    cod5: float = as2 * pro5_strength
    cod6: float = pro6_strength * (as3 + as4)
    cod7: float = pro7_strength * (as7 + as8)
    cod8: float = as5 * pro8_strength * rs4
    cod9: float = as6 * pro9_strength * rs5
    pp9_v1: float = cod9
    pp8_v1: float = cod8
    pp7_v1: float = cod7
    pp6_v1: float = cod6
    pp5_v1: float = cod5
    pp4_v1: float = cod4
    pp3_v1: float = cod3
    pp2_v1: float = cod2
    return {
        "as8": as8,
        "cod1": cod1,
        "rs1": rs1,
        "rs2": rs2,
        "rs3": rs3,
        "rs4": rs4,
        "rs5": rs5,
        "rs6": rs6,
        "rs7": rs7,
        "as1": as1,
        "as2": as2,
        "as3": as3,
        "as4": as4,
        "as5": as5,
        "as6": as6,
        "as7": as7,
        "pp9_mrna_conc": pp9_mrna_conc,
        "p9_conc": p9_conc,
        "pp8_mrna_conc": pp8_mrna_conc,
        "p8_conc": p8_conc,
        "pp7_mrna_conc": pp7_mrna_conc,
        "p7_conc": p7_conc,
        "pp6_mrna_conc": pp6_mrna_conc,
        "p6_conc": p6_conc,
        "pp5_mrna_conc": pp5_mrna_conc,
        "p5_conc": p5_conc,
        "pp4_mrna_conc": pp4_mrna_conc,
        "p4_conc": p4_conc,
        "pp3_mrna_conc": pp3_mrna_conc,
        "p3_conc": p3_conc,
        "pp2_mrna_conc": pp2_mrna_conc,
        "p2_conc": p2_conc,
        "pp1_mrna_conc": pp1_mrna_conc,
        "p1_conc": p1_conc,
        "pp9_v2": pp9_v2,
        "pp9_v3": pp9_v3,
        "pp9_v4": pp9_v4,
        "pp8_v2": pp8_v2,
        "pp8_v3": pp8_v3,
        "pp8_v4": pp8_v4,
        "pp7_v2": pp7_v2,
        "pp7_v3": pp7_v3,
        "pp7_v4": pp7_v4,
        "pp6_v2": pp6_v2,
        "pp6_v3": pp6_v3,
        "pp6_v4": pp6_v4,
        "pp5_v2": pp5_v2,
        "pp5_v3": pp5_v3,
        "pp5_v4": pp5_v4,
        "pp4_v2": pp4_v2,
        "pp4_v3": pp4_v3,
        "pp4_v4": pp4_v4,
        "pp3_v2": pp3_v2,
        "pp3_v3": pp3_v3,
        "pp3_v4": pp3_v4,
        "pp2_v2": pp2_v2,
        "pp2_v3": pp2_v3,
        "pp2_v4": pp2_v4,
        "pp1_v1": pp1_v1,
        "pp1_v2": pp1_v2,
        "pp1_v3": pp1_v3,
        "pp1_v4": pp1_v4,
        "cod2": cod2,
        "cod3": cod3,
        "cod4": cod4,
        "cod5": cod5,
        "cod6": cod6,
        "cod7": cod7,
        "cod8": cod8,
        "cod9": cod9,
        "pp9_v1": pp9_v1,
        "pp8_v1": pp8_v1,
        "pp7_v1": pp7_v1,
        "pp6_v1": pp6_v1,
        "pp5_v1": pp5_v1,
        "pp4_v1": pp4_v1,
        "pp3_v1": pp3_v1,
        "pp2_v1": pp2_v1,
    }



import math

time: float = 0.0
k1: float = 1.02300000000000
compartment: float = math.nan
S2: float = 0.0
S1: float = 0.150000000000000

# Initial assignments
reaction1 = S1 * k1
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    reaction1: float = S1 * k1
    dS1dt: float = -reaction1
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    reaction1: float = S1 * k1
    return {
        "reaction1": reaction1,
    }



time: float = 0.0
c: float = 0.500000000000000

# Initial assignments
dc = 0.5 * c
y0 = [c]
variable_names = ["c"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (c,) = variables
    dc: float = 0.5 * c
    dcdt: float = dc
    return (dcdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (c,) = variables
    dc: float = 0.5 * c
    return {
        "dc": dc,
    }



time: float = 0.0
kf: float = 1.10000000000000
kr: float = 0.0900000000000000
C: float = 1.00000000000000
S3: float = 0.400000000000000
S1: float = 1.00000000000000
S2: float = 0.500000000000000

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
S3_conc = S3 / C
reaction1 = -S1_conc * S2_conc * kf + S3_conc * kr
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    reaction1: float = -S1_conc * S2_conc * kf + S3_conc * kr
    dS1dt: float = reaction1
    dS2dt: float = reaction1
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    reaction1: float = -S1_conc * S2_conc * kf + S3_conc * kr
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "S3_conc": S3_conc,
        "reaction1": reaction1,
    }



time: float = 0.0
S1: float = 1.00000000000000
C: float = 1.00000000000000

# Initial assignments
dC = 1
y0 = [C]
variable_names = ["C"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (C,) = variables
    dC: float = 1
    dCdt: float = dC
    return (dCdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (C,) = variables
    dC: float = 1
    return {
        "dC": dC,
    }



time: float = 0.0
C: float = 1.00000000000000
A: float = 1.00000000000000

# Initial assignments
dA1_sr2 = -1
dA1_sr = 1
J0 = 1
A1_sr = 2
A1_sr2 = 1
y0 = [A, A1_sr, A1_sr2]
variable_names = ["A", "A1_sr", "A1_sr2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    A, A1_sr, A1_sr2 = variables
    dA1_sr2: float = -1
    dA1_sr: float = 1
    J0: float = 1
    dA1_sr2dt: float = dA1_sr2
    dA1_srdt: float = dA1_sr
    dAdt: float = J0 * (A1_sr + A1_sr2)
    return dAdt, dA1_srdt, dA1_sr2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    A, A1_sr, A1_sr2 = variables
    dA1_sr2: float = -1
    dA1_sr: float = 1
    J0: float = 1
    return {
        "dA1_sr2": dA1_sr2,
        "dA1_sr": dA1_sr,
        "J0": J0,
    }



time: float = 0.0
k1: float = 0.750000000000000
k2: float = 50.0000000000000

# Initial assignments
S1 = 1.33333333333333 * k1
S2 = 3.0e-17 * k2
S3 = S2 * k1
dS1 = -S1 * k2
dS2 = S1 * k2
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    dS1: float = -S1 * k2
    dS2: float = S1 * k2
    dS1dt: float = dS1
    dS2dt: float = dS2
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    S3: float = S2 * k1
    dS1: float = -S1 * k2
    dS2: float = S1 * k2
    return {
        "S3": S3,
        "dS1": dS1,
        "dS2": dS2,
    }



time: float = 0.0
p1: float = 1.00000000000000
p2: float = 2.00000000000000

# Initial assignments
dp1 = True
dp2 = False
y0 = [p1, p2]
variable_names = ["p1", "p2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    p1, p2 = variables
    dp1: float = True
    dp2: float = False
    dp1dt: float = dp1
    dp2dt: float = dp2
    return dp1dt, dp2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    p1, p2 = variables
    dp1: float = True
    dp2: float = False
    return {
        "dp1": dp1,
        "dp2": dp2,
    }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import wadler_lindig as wl

from pysbml.parse.data import md_table_from_dict

__all__ = ["Expr", "Model", "Parameter", "Reaction", "Stoichiometry", "Variable"]

if TYPE_CHECKING:
    import sympy
    from sympy.physics.units.quantities import Quantity


def _md_eq(s: Expr) -> str:
    return f"${s}$".replace("|", r"\|").replace("&", r"\&").replace("_", r"\_")


type Expr = sympy.Symbol | sympy.Float | sympy.Expr
type Stoichiometry = dict[str, Expr]


@dataclass(kw_only=True, slots=True)
class Parameter:
    value: sympy.Float
    unit: Quantity | None

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Variable:
    value: sympy.Float
    unit: Quantity | None

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Reaction:
    expr: sympy.Expr
    stoichiometry: Stoichiometry

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Model:
    name: str
    units: dict[str, Quantity] = field(default_factory=dict)
    functions: dict[str, Expr] = field(default_factory=dict)
    parameters: dict[str, Parameter] = field(default_factory=dict)
    variables: dict[str, Variable] = field(default_factory=dict)
    derived: dict[str, Expr] = field(default_factory=dict)
    reactions: dict[str, Reaction] = field(default_factory=dict)
    initial_assignments: dict[str, Expr] = field(default_factory=dict)

    def __repr__(self) -> str:
        return wl.pformat(self)

    def _repr_markdown_(self) -> str:
        content = [f"# {self.name}"]

        if len(self.functions) > 0:
            content.append("# Functions")
            content.append(
                md_table_from_dict(
                    headers=[
                        "name",
                        "body",
                    ],
                    els=[(k, _md_eq(v)) for k, v in self.functions.items()],
                )
            )

        if len(self.parameters) > 0:
            content.append("# Parameters")
            content.append(
                md_table_from_dict(
                    headers=[
                        "name",
                        "value",
                        "unit",
                    ],
                    els=[(k, v.value, v.unit) for k, v in self.parameters.items()],
                )
            )
        if len(self.variables) > 0:
            content.append("# Variables")
            content.append(
                md_table_from_dict(
                    headers=[
                        "name",
                        "value",
                        "unit",
                    ],
                    els=[
                        (
                            k,
                            _md_eq(v.value),
                            v.unit,
                        )
                        for k, v in self.variables.items()
                    ],
                )
            )
        if len(self.derived) > 0:
            content.append("# Derived")
            content.append(
                md_table_from_dict(
                    headers=["name", "fn"],
                    els=[(k, _md_eq(v)) for k, v in self.derived.items()],
                )
            )
        if len(self.initial_assignments) > 0:
            content.append("# Initial assignments")
            content.append(
                md_table_from_dict(
                    headers=["name", "fn"],
                    els=[(k, _md_eq(v)) for k, v in self.initial_assignments.items()],
                )
            )

        if len(self.reactions) > 0:
            content.append("# Reactions")
            content.append(
                md_table_from_dict(
                    headers=["name", "fn", "stoichiometry"],
                    els=[
                        (k, _md_eq(v.expr), v.stoichiometry)
                        for k, v in self.reactions.items()
                    ],
                )
            )
        return "\n".join(content)



time: float = 0.0
c: float = 1.00000000000000
S1: float = 2.00000000000000

# Initial assignments
S1_conc = S1 / c
J0 = 2
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_conc: float = S1 / c
    J0: float = 2
    return {
        "S1_conc": S1_conc,
        "J0": J0,
    }



time: float = 0.0

# Initial assignments
a = 3
b = 4
c = 5
d = 6
e = 7
f = 8
g = 9
h = 10
i = 11
j = 12
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    a: float = 3
    b: float = 4
    c: float = 5
    d: float = 6
    e: float = 7
    f: float = 8
    g: float = 9
    h: float = 10
    i: float = 11
    j: float = 12
    return {
        "a": a,
        "b": b,
        "c": c,
        "d": d,
        "e": e,
        "f": f,
        "g": g,
        "h": h,
        "i": i,
        "j": j,
    }



time: float = 0.0
x: float = 3.00000000000000

# Initial assignments
dx = 0
y0 = [x]
variable_names = ["x"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (x,) = variables
    dx: float = 0
    dxdt: float = dx
    return (dxdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (x,) = variables
    dx: float = 0
    return {
        "dx": dx,
    }



time: float = 0.0
k1: float = 1.50000000000000
compartment: float = 1.00000000000000
S1: float = 1.50000000000000e-6
S2: float = 1.00000000000000e-6

# Initial assignments
S1_conc = S1 / compartment
S2_conc = S2 / compartment
reaction1 = S1_conc * compartment * k1
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_conc: float = S1 / compartment
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    reaction1: float = S1_conc * compartment * k1
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "reaction1": reaction1,
    }



import math

time: float = 0.0
C: float = 1.00000000000000
S1_stoich: float = 2.00000000000000
S1: float = 0.0

# Initial assignments
P1 = math.exp(S1_stoich)
P2 = abs(S1_stoich)
P3 = abs(S1_stoich)
P4 = math.acos(1 - S1_stoich)
P5 = math.acos((1 / 4) * S1_stoich)
P6 = math.asin(S1_stoich - 1)
P7 = -math.asin((1 / 4) * S1_stoich)
P8 = math.atan(S1_stoich + 0.8)
P9 = -math.atan(3 * S1_stoich + 1.09)
P10 = math.ceil((1 / 4) * S1_stoich)
P11 = math.ceil(4 * S1_stoich - 0.45)
P12 = math.ceil(-2 * S1_stoich - 0.6)
P13 = math.cos(4 * S1_stoich + 1.1)
P14 = math.cos((1 / 10) * S1_stoich + 0.02)
P15 = 1
P16 = math.exp((1 / 2) * S1_stoich)
P17 = 0.718923733431926 * math.exp((1 / 2) * S1_stoich)
P18 = math.floor(-2 * S1_stoich - 0.6)
P19 = math.floor(4 * S1_stoich + 1.1)
P20 = math.log((1 / 10) * S1_stoich)
P21 = math.log((1 / 2) * S1_stoich)
P22 = math.log((1 / 10) * S1_stoich) / math.log(10)
P23 = math.log((1 / 2) * S1_stoich) / math.log(10)
P24 = 1
P25 = S1_stoich**S1_stoich
P26 = 5.56228172775440
P27 = (S1_stoich**2) ** S1_stoich
P28 = 3.1**S1_stoich
P29 = math.sqrt(2) * math.sqrt(S1_stoich)
P30 = math.sqrt((1 / 5) * S1_stoich + 7)
P31 = math.sin(S1_stoich + 0.1)
P32 = 0
P33 = -math.sin(2 * S1_stoich + 1.9)
P34 = 0
P35 = math.tan((1 / 2) * S1_stoich + 0.11)
P36 = -math.tan(3 * S1_stoich)
P37 = 1 / math.cos((1 / 4) * S1_stoich)
P38 = 1 / math.sin(2.25 * S1_stoich)
P39 = 1 / math.tan((1 / 10) * S1_stoich)
P40 = math.sinh((1 / 10) * S1_stoich + 0.1)
P41 = math.cosh(S1_stoich - 0.3)
P42 = math.acos(1 / (S1_stoich + 0.3))
P43 = math.asin(1 / (S1_stoich - 0.9))
P44 = math.atan(1 / (S1_stoich - 2.1))
P45 = math.asinh(50 * S1_stoich - 1)
P46 = math.acosh((1 / 2) * S1_stoich + 0.34)
P47 = math.atanh(S1_stoich - 2.7)
P48 = math.log(
    math.sqrt(-1 + 4.76190476190476 / S1_stoich)
    * math.sqrt(1 + 4.76190476190476 / S1_stoich)
    + 4.76190476190476 / S1_stoich
)
P49 = math.log(math.sqrt(1 + 40000.0 / S1_stoich**2) + 200.0 / S1_stoich)
S1_conc = S1 / C
J0 = 1
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    J0: float = 1
    dS1dt: float = C * J0 * S1_stoich
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    P1: float = math.exp(S1_stoich)
    P2: float = abs(S1_stoich)
    P3: float = abs(S1_stoich)
    P4: float = math.acos(1 - S1_stoich)
    P5: float = math.acos((1 / 4) * S1_stoich)
    P6: float = math.asin(S1_stoich - 1)
    P7: float = -math.asin((1 / 4) * S1_stoich)
    P8: float = math.atan(S1_stoich + 0.8)
    P9: float = -math.atan(3 * S1_stoich + 1.09)
    P10: float = math.ceil((1 / 4) * S1_stoich)
    P11: float = math.ceil(4 * S1_stoich - 0.45)
    P12: float = math.ceil(-2 * S1_stoich - 0.6)
    P13: float = math.cos(4 * S1_stoich + 1.1)
    P14: float = math.cos((1 / 10) * S1_stoich + 0.02)
    P15: float = 1
    P16: float = math.exp((1 / 2) * S1_stoich)
    P17: float = 0.718923733431926 * math.exp((1 / 2) * S1_stoich)
    P18: float = math.floor(-2 * S1_stoich - 0.6)
    P19: float = math.floor(4 * S1_stoich + 1.1)
    P20: float = math.log((1 / 10) * S1_stoich)
    P21: float = math.log((1 / 2) * S1_stoich)
    P22: float = math.log((1 / 10) * S1_stoich) / math.log(10)
    P23: float = math.log((1 / 2) * S1_stoich) / math.log(10)
    P24: float = 1
    P25: float = S1_stoich**S1_stoich
    P26: float = 5.56228172775440
    P27: float = (S1_stoich**2) ** S1_stoich
    P28: float = 3.1**S1_stoich
    P29: float = math.sqrt(2) * math.sqrt(S1_stoich)
    P30: float = math.sqrt((1 / 5) * S1_stoich + 7)
    P31: float = math.sin(S1_stoich + 0.1)
    P32: float = 0
    P33: float = -math.sin(2 * S1_stoich + 1.9)
    P34: float = 0
    P35: float = math.tan((1 / 2) * S1_stoich + 0.11)
    P36: float = -math.tan(3 * S1_stoich)
    P37: float = 1 / math.cos((1 / 4) * S1_stoich)
    P38: float = 1 / math.sin(2.25 * S1_stoich)
    P39: float = 1 / math.tan((1 / 10) * S1_stoich)
    P40: float = math.sinh((1 / 10) * S1_stoich + 0.1)
    P41: float = math.cosh(S1_stoich - 0.3)
    P42: float = math.acos(1 / (S1_stoich + 0.3))
    P43: float = math.asin(1 / (S1_stoich - 0.9))
    P44: float = math.atan(1 / (S1_stoich - 2.1))
    P45: float = math.asinh(50 * S1_stoich - 1)
    P46: float = math.acosh((1 / 2) * S1_stoich + 0.34)
    P47: float = math.atanh(S1_stoich - 2.7)
    P48: float = math.log(
        math.sqrt(-1 + 4.76190476190476 / S1_stoich)
        * math.sqrt(1 + 4.76190476190476 / S1_stoich)
        + 4.76190476190476 / S1_stoich
    )
    P49: float = math.log(math.sqrt(1 + 40000.0 / S1_stoich**2) + 200.0 / S1_stoich)
    S1_conc: float = S1 / C
    J0: float = 1
    return {
        "P1": P1,
        "P2": P2,
        "P3": P3,
        "P4": P4,
        "P5": P5,
        "P6": P6,
        "P7": P7,
        "P8": P8,
        "P9": P9,
        "P10": P10,
        "P11": P11,
        "P12": P12,
        "P13": P13,
        "P14": P14,
        "P15": P15,
        "P16": P16,
        "P17": P17,
        "P18": P18,
        "P19": P19,
        "P20": P20,
        "P21": P21,
        "P22": P22,
        "P23": P23,
        "P24": P24,
        "P25": P25,
        "P26": P26,
        "P27": P27,
        "P28": P28,
        "P29": P29,
        "P30": P30,
        "P31": P31,
        "P32": P32,
        "P33": P33,
        "P34": P34,
        "P35": P35,
        "P36": P36,
        "P37": P37,
        "P38": P38,
        "P39": P39,
        "P40": P40,
        "P41": P41,
        "P42": P42,
        "P43": P43,
        "P44": P44,
        "P45": P45,
        "P46": P46,
        "P47": P47,
        "P48": P48,
        "P49": P49,
        "S1_conc": S1_conc,
        "J0": J0,
    }



from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import libsbml

from .name_conversion import name_to_py
from .units import get_ast_types

LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    from libsbml import ASTNode


@dataclass
class Base: ...


@dataclass
class Symbol(Base):
    name: str

    def __repr__(self) -> str:
        return self.name


@dataclass
class Constant(Base):
    name: str

    def __repr__(self) -> str:
        return self.name


@dataclass
class Boolean(Base):
    value: bool


@dataclass
class Integer(Base):
    value: int

    def __repr__(self) -> str:
        return f"{self.value}"


@dataclass
class Float(Base):
    value: float

    def __repr__(self) -> str:
        return f"{self.value:.2g}"


###############################################################################
# Unary fns
###############################################################################


@dataclass
class Abs(Base):
    child: Base


@dataclass
class Ceiling(Base):
    child: Base


@dataclass
class Exp(Base):
    child: Base


@dataclass
class Factorial(Base):
    child: Base


@dataclass
class Floor(Base):
    child: Base


@dataclass
class Ln(Base):
    child: Base


@dataclass
class Log(Base):
    base: Base
    child: Base


@dataclass
class Sqrt(Base):
    base: Base
    child: Base


@dataclass
class Sin(Base):
    child: Base


@dataclass
class Cos(Base):
    child: Base


@dataclass
class Tan(Base):
    child: Base


@dataclass
class Sec(Base):
    child: Base


@dataclass
class Csc(Base):
    child: Base


@dataclass
class Cot(Base):
    child: Base


@dataclass
class Asin(Base):
    child: Base


@dataclass
class Acos(Base):
    child: Base


@dataclass
class Atan(Base):
    child: Base


@dataclass
class Acot(Base):
    child: Base


@dataclass
class ArcSec(Base):
    child: Base


@dataclass
class ArcCsc(Base):
    child: Base


@dataclass
class Sinh(Base):
    child: Base


@dataclass
class Cosh(Base):
    child: Base


@dataclass
class Tanh(Base):
    child: Base


@dataclass
class Sech(Base):
    child: Base


@dataclass
class Csch(Base):
    child: Base


@dataclass
class Coth(Base):
    child: Base


@dataclass
class ArcSinh(Base):
    child: Base


@dataclass
class ArcCosh(Base):
    child: Base


@dataclass
class ArcTanh(Base):
    child: Base


@dataclass
class ArcCsch(Base):
    child: Base


@dataclass
class ArcSech(Base):
    child: Base


@dataclass
class ArcCoth(Base):
    child: Base


@dataclass
class RateOf(Base):
    target: Base


###############################################################################
# Binary fns
###############################################################################


@dataclass
class Pow(Base):
    left: Base
    right: Base

    def __repr__(self) -> str:
        return f"{self.left!r} ** {self.right!r}"


@dataclass
class Implies(Base):
    left: Base
    right: Base


###############################################################################
# n-ary fns
###############################################################################


@dataclass
class Function(Base):
    name: str
    children: list[Base]


@dataclass
class Max(Base):
    children: list[Base]


@dataclass
class Min(Base):
    children: list[Base]


@dataclass
class Piecewise(Base):
    children: list[Base]


@dataclass
class Rem(Base):
    children: list[Base]


@dataclass
class Lambda(Base):
    fn: Base
    args: list[Base]


@dataclass
class And(Base):
    children: list[Base]


@dataclass
class Not(Base):
    children: list[Base]


@dataclass
class Or(Base):
    children: list[Base]


@dataclass
class Xor(Base):
    children: list[Base]


@dataclass
class Eq(Base):
    children: list[Base]


@dataclass
class GreaterEqual(Base):
    children: list[Base]


@dataclass
class GreaterThan(Base):
    children: list[Base]


@dataclass
class LessEqual(Base):
    children: list[Base]


@dataclass
class LessThan(Base):
    """
    a < b
    a < b < c
    """

    children: list[Base]


@dataclass
class NotEqual(Base):
    children: list[Base]


@dataclass
class Add(Base):
    children: list[Base]


@dataclass
class Minus(Base):
    children: list[Base]


@dataclass
class Mul(Base):
    children: list[Base]

    def __repr__(self) -> str:
        return " * ".join(repr(i) for i in self.children)


@dataclass
class Divide(Base):
    children: list[Base]


@dataclass
class IntDivide(Base):
    children: list[Base]


@dataclass
class Delay(Base):
    children: list[Base]


__all__ = [
    "AST_TYPES",
    "Abs",
    "Acos",
    "Acot",
    "Add",
    "And",
    "ArcCosh",
    "ArcCoth",
    "ArcCsc",
    "ArcCsch",
    "ArcSec",
    "ArcSech",
    "ArcSinh",
    "ArcTanh",
    "Asin",
    "Atan",
    "Base",
    "Boolean",
    "Ceiling",
    "Constant",
    "Cos",
    "Cosh",
    "Cot",
    "Coth",
    "Csc",
    "Csch",
    "Delay",
    "Divide",
    "Eq",
    "Exp",
    "Factorial",
    "Float",
    "Floor",
    "Function",
    "GreaterEqual",
    "GreaterThan",
    "Implies",
    "IntDivide",
    "Integer",
    "LOGGER",
    "Lambda",
    "LessEqual",
    "LessThan",
    "Ln",
    "Log",
    "Max",
    "Min",
    "Minus",
    "Mul",
    "Not",
    "NotEqual",
    "Or",
    "Piecewise",
    "Pow",
    "RateOf",
    "Rem",
    "Sec",
    "Sech",
    "Sin",
    "Sinh",
    "Sqrt",
    "Symbol",
    "Tan",
    "Tanh",
    "Xor",
    "handle_ast_constant_e",
    "handle_ast_constant_false",
    "handle_ast_constant_pi",
    "handle_ast_constant_true",
    "handle_ast_divide",
    "handle_ast_divide_int",
    "handle_ast_function",
    "handle_ast_function_abs",
    "handle_ast_function_ceiling",
    "handle_ast_function_delay",
    "handle_ast_function_exp",
    "handle_ast_function_factorial",
    "handle_ast_function_floor",
    "handle_ast_function_ln",
    "handle_ast_function_log",
    "handle_ast_function_max",
    "handle_ast_function_min",
    "handle_ast_function_piecewise",
    "handle_ast_function_power",
    "handle_ast_function_rate_of",
    "handle_ast_function_rem",
    "handle_ast_function_root",
    "handle_ast_integer",
    "handle_ast_lambda",
    "handle_ast_logical_and",
    "handle_ast_logical_implies",
    "handle_ast_logical_not",
    "handle_ast_logical_or",
    "handle_ast_logical_xor",
    "handle_ast_minus",
    "handle_ast_name",
    "handle_ast_name_avogadro",
    "handle_ast_name_time",
    "handle_ast_originates_in_package",
    "handle_ast_plus",
    "handle_ast_rational",
    "handle_ast_real",
    "handle_ast_relational_eq",
    "handle_ast_relational_geq",
    "handle_ast_relational_gt",
    "handle_ast_relational_leq",
    "handle_ast_relational_lt",
    "handle_ast_relational_neq",
    "handle_ast_times",
    "handle_ast_trigonometric_arc_cos",
    "handle_ast_trigonometric_arc_cosh",
    "handle_ast_trigonometric_arc_cot",
    "handle_ast_trigonometric_arc_coth",
    "handle_ast_trigonometric_arc_csc",
    "handle_ast_trigonometric_arc_csch",
    "handle_ast_trigonometric_arc_sec",
    "handle_ast_trigonometric_arc_sech",
    "handle_ast_trigonometric_arc_sin",
    "handle_ast_trigonometric_arc_sinh",
    "handle_ast_trigonometric_arc_tan",
    "handle_ast_trigonometric_arc_tanh",
    "handle_ast_trigonometric_cos",
    "handle_ast_trigonometric_cosh",
    "handle_ast_trigonometric_cot",
    "handle_ast_trigonometric_coth",
    "handle_ast_trigonometric_csc",
    "handle_ast_trigonometric_csch",
    "handle_ast_trigonometric_sec",
    "handle_ast_trigonometric_sech",
    "handle_ast_trigonometric_sin",
    "handle_ast_trigonometric_sinh",
    "handle_ast_trigonometric_tan",
    "handle_ast_trigonometric_tanh",
    "parse_sbml_math",
]


AST_TYPES = get_ast_types()


def handle_ast_constant_e(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    return Constant("e")


def handle_ast_constant_false(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    return Boolean(value=False)


def handle_ast_constant_true(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    return Boolean(value=True)


def handle_ast_constant_pi(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    return Constant("pi")


def handle_ast_function(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    name = node.getName()
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("name: %s, args: %s", name, children)
    return Function(name=name, children=children)


def handle_ast_function_abs(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Abs(child=child)


def handle_ast_function_ceiling(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Ceiling(child=child)


def handle_ast_function_delay(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    return Delay(children=children)


def handle_ast_function_exp(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Exp(child=child)


def handle_ast_function_factorial(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Factorial(child=child)


def handle_ast_function_floor(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Floor(child=child)


def handle_ast_function_ln(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Ln(child=child)


def handle_ast_function_log(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    base = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    arg = _handle_ast_node(node=node.getChild(1), func_arguments=func_arguments)
    LOGGER.debug("base: %s, child: %s", base, arg)
    return Log(base=base, child=arg)


def handle_ast_function_max(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Max(children=children)


def handle_ast_function_min(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Min(children=children)


def handle_ast_function_piecewise(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    """
    <piecewise>
    <piece> value condition </piece>
        ....
    <piece> value condition </piece>
    <otherwise> value </otherwise>
    </piecewise>
    """
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Piecewise(children=children)


def handle_ast_function_power(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Pow(left=children[0], right=children[1])


def handle_ast_function_rate_of(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return RateOf(target=child)


def handle_ast_function_root(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    base = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    child = _handle_ast_node(node=node.getChild(1), func_arguments=func_arguments)
    LOGGER.debug("child: %s", child)
    return Sqrt(base=base, child=child)


def handle_ast_function_rem(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return Rem(children=children)


def handle_ast_integer(
    node: ASTNode,
    func_arguments: list[Symbol],  # noqa: ARG001,
) -> Base:
    child = node.getValue()
    LOGGER.debug("child: %s", child)
    return Integer(value=child)


def handle_ast_lambda(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    """
    A function having n variables is specified using a lambda element with n + 1 child elements.
    Each of the first n child elements is a bvar element that represents one of the variables of the function
    The (n + 1)th element is an expression that defines the function itself
    """
    # num_b_vars = node.getNumBvars()
    num_children = node.getNumChildren()
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(num_children)
    ]
    LOGGER.debug("children: %s", children)
    return Lambda(args=children[:-1], fn=children[-1])


def handle_ast_logical_and(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return And(children=children)


def handle_ast_logical_implies(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    return Implies(children[0], children[1])


def handle_ast_logical_not(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Not(children=children)


def handle_ast_logical_or(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Or(children=children)


def handle_ast_logical_xor(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Xor(children=children)


def handle_ast_plus(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Add(children=children)


def handle_ast_minus(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Minus(children=children)


def handle_ast_times(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Mul(children=children)


def handle_ast_divide(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Divide(children=children)


def handle_ast_divide_int(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return IntDivide(children=children)


def handle_ast_name(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    name = Symbol(name_to_py(node.getName()))
    if name not in func_arguments:
        func_arguments.append(name)
    LOGGER.debug("name: %s", name)
    return name


def handle_ast_name_avogadro(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],  # noqa: ARG001,
) -> Base:
    return Float(6.02214179e23)


def handle_ast_name_time(
    node: ASTNode,  # noqa: ARG001
    func_arguments: list[Symbol],
) -> Base:
    name = Symbol("time")
    if name not in func_arguments:
        func_arguments.append(name)
    return name


def handle_ast_originates_in_package(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    raise NotImplementedError


def handle_ast_rational(
    node: ASTNode,
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    val = node.getValue()
    LOGGER.debug("val: %s", val)
    return Float(value=val)


def handle_ast_real(
    node: ASTNode,
    func_arguments: list[Symbol],  # noqa: ARG001
) -> Base:
    value = str(node.getValue())
    LOGGER.debug("value: %s", value)
    if value == "inf":
        return Float(math.inf)
    if value == "nan":
        return Float(math.nan)

    # FIXME: seems dumb
    return Float(value=float(value))


def handle_ast_relational_eq(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return Eq(children=children)


def handle_ast_relational_geq(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return GreaterEqual(children=children)


def handle_ast_relational_gt(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return GreaterThan(children=children)


def handle_ast_relational_leq(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return LessEqual(children=children)


def handle_ast_relational_lt(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]
    LOGGER.debug("children: %s", children)
    return LessThan(children=children)


def handle_ast_relational_neq(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    children = [
        _handle_ast_node(node=node.getChild(i), func_arguments=func_arguments)
        for i in range(node.getNumChildren())
    ]

    LOGGER.debug("children: %s", children)
    return NotEqual(children=children)


###############################################################################
# Base
###############################################################################


def handle_ast_trigonometric_sin(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Sin(child=child)


def handle_ast_trigonometric_cos(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Cos(child=child)


def handle_ast_trigonometric_tan(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Tan(child=child)


def handle_ast_trigonometric_sec(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Sec(child)


def handle_ast_trigonometric_csc(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Csc(child)


def handle_ast_trigonometric_cot(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Cot(child)


###############################################################################
# Inverse
###############################################################################


def handle_ast_trigonometric_arc_sin(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Asin(child)


def handle_ast_trigonometric_arc_cos(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Acos(child)


def handle_ast_trigonometric_arc_tan(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Atan(child)


def handle_ast_trigonometric_arc_cot(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Acot(child)


def handle_ast_trigonometric_arc_sec(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcSec(child)


def handle_ast_trigonometric_arc_csc(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcCsc(child)


###############################################################################
# Hyperbolic
###############################################################################


def handle_ast_trigonometric_sinh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Sinh(child)


def handle_ast_trigonometric_cosh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Cosh(child)


def handle_ast_trigonometric_tanh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Tanh(child)


def handle_ast_trigonometric_sech(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Sech(child)


def handle_ast_trigonometric_csch(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Csch(child)


def handle_ast_trigonometric_coth(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return Coth(child)


###############################################################################
# Hyperbolic - inverse
###############################################################################


def handle_ast_trigonometric_arc_sinh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcSinh(child)


def handle_ast_trigonometric_arc_cosh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcCosh(child)


def handle_ast_trigonometric_arc_tanh(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcTanh(child)


def handle_ast_trigonometric_arc_csch(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcCsch(child)


def handle_ast_trigonometric_arc_sech(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)
    return ArcSech(child)


def handle_ast_trigonometric_arc_coth(
    node: ASTNode,
    func_arguments: list[Symbol],
) -> Base:
    child = _handle_ast_node(node=node.getChild(0), func_arguments=func_arguments)

    return ArcCoth(child)


def _handle_ast_node(node: ASTNode, func_arguments: list[Symbol]) -> Base:
    commands = {
        "AST_CONSTANT_E": handle_ast_constant_e,
        "AST_CONSTANT_FALSE": handle_ast_constant_false,
        "AST_CONSTANT_PI": handle_ast_constant_pi,
        "AST_CONSTANT_TRUE": handle_ast_constant_true,
        "AST_DIVIDE": handle_ast_divide,
        "AST_FUNCTION": handle_ast_function,
        "AST_FUNCTION_ABS": handle_ast_function_abs,
        "AST_FUNCTION_ARCCOS": handle_ast_trigonometric_arc_cos,
        "AST_FUNCTION_ARCCOSH": handle_ast_trigonometric_arc_cosh,
        "AST_FUNCTION_ARCCOT": handle_ast_trigonometric_arc_cot,
        "AST_FUNCTION_ARCCOTH": handle_ast_trigonometric_arc_coth,
        "AST_FUNCTION_ARCCSC": handle_ast_trigonometric_arc_csc,
        "AST_FUNCTION_ARCCSCH": handle_ast_trigonometric_arc_csch,
        "AST_FUNCTION_ARCSEC": handle_ast_trigonometric_arc_sec,
        "AST_FUNCTION_ARCSECH": handle_ast_trigonometric_arc_sech,
        "AST_FUNCTION_ARCSIN": handle_ast_trigonometric_arc_sin,
        "AST_FUNCTION_ARCSINH": handle_ast_trigonometric_arc_sinh,
        "AST_FUNCTION_ARCTAN": handle_ast_trigonometric_arc_tan,
        "AST_FUNCTION_ARCTANH": handle_ast_trigonometric_arc_tanh,
        "AST_FUNCTION_CEILING": handle_ast_function_ceiling,
        "AST_FUNCTION_COS": handle_ast_trigonometric_cos,
        "AST_FUNCTION_COSH": handle_ast_trigonometric_cosh,
        "AST_FUNCTION_COT": handle_ast_trigonometric_cot,
        "AST_FUNCTION_COTH": handle_ast_trigonometric_coth,
        "AST_FUNCTION_CSC": handle_ast_trigonometric_csc,
        "AST_FUNCTION_CSCH": handle_ast_trigonometric_csch,
        "AST_FUNCTION_DELAY": handle_ast_function_delay,
        "AST_FUNCTION_EXP": handle_ast_function_exp,
        "AST_FUNCTION_FACTORIAL": handle_ast_function_factorial,
        "AST_FUNCTION_FLOOR": handle_ast_function_floor,
        "AST_FUNCTION_LN": handle_ast_function_ln,
        "AST_FUNCTION_LOG": handle_ast_function_log,
        "AST_FUNCTION_MAX": handle_ast_function_max,
        "AST_FUNCTION_MIN": handle_ast_function_min,
        "AST_FUNCTION_PIECEWISE": handle_ast_function_piecewise,
        "AST_FUNCTION_POWER": handle_ast_function_power,
        "AST_FUNCTION_QUOTIENT": handle_ast_divide_int,
        "AST_FUNCTION_RATE_OF": handle_ast_function_rate_of,
        "AST_FUNCTION_ROOT": handle_ast_function_root,
        "AST_FUNCTION_REM": handle_ast_function_rem,
        "AST_FUNCTION_SEC": handle_ast_trigonometric_sec,
        "AST_FUNCTION_SECH": handle_ast_trigonometric_sech,
        "AST_FUNCTION_SIN": handle_ast_trigonometric_sin,
        "AST_FUNCTION_SINH": handle_ast_trigonometric_sinh,
        "AST_FUNCTION_TAN": handle_ast_trigonometric_tan,
        "AST_FUNCTION_TANH": handle_ast_trigonometric_tanh,
        "AST_INTEGER": handle_ast_integer,
        "AST_LAMBDA": handle_ast_lambda,
        "AST_LOGICAL_AND": handle_ast_logical_and,
        "AST_LOGICAL_IMPLIES": handle_ast_logical_implies,
        "AST_LOGICAL_NOT": handle_ast_logical_not,
        "AST_LOGICAL_OR": handle_ast_logical_or,
        "AST_LOGICAL_XOR": handle_ast_logical_xor,
        "AST_MINUS": handle_ast_minus,
        "AST_NAME": handle_ast_name,
        "AST_NAME_AVOGADRO": handle_ast_name_avogadro,
        "AST_NAME_TIME": handle_ast_name_time,
        "AST_ORIGINATES_IN_PACKAGE": handle_ast_originates_in_package,
        "AST_PLUS": handle_ast_plus,
        "AST_POWER": handle_ast_function_power,
        "AST_RATIONAL": handle_ast_rational,
        "AST_REAL": handle_ast_real,
        "AST_REAL_E": handle_ast_real,
        "AST_RELATIONAL_EQ": handle_ast_relational_eq,
        "AST_RELATIONAL_GEQ": handle_ast_relational_geq,
        "AST_RELATIONAL_GT": handle_ast_relational_gt,
        "AST_RELATIONAL_LEQ": handle_ast_relational_leq,
        "AST_RELATIONAL_LT": handle_ast_relational_lt,
        "AST_RELATIONAL_NEQ": handle_ast_relational_neq,
        "AST_TIMES": handle_ast_times,
    }
    return commands[AST_TYPES[node.getType()]](node=node, func_arguments=func_arguments)


def parse_sbml_math(node: ASTNode) -> tuple[Base, list[Symbol]]:
    func_arguments: list[Any] = []
    try:
        body = _handle_ast_node(node=node, func_arguments=func_arguments)
    except TypeError as e:
        msg = f"Cannot parse rule {libsbml.formulaToL3String(node)}"
        raise TypeError(msg) from e

    LOGGER.debug("%s", func_arguments)
    return body, func_arguments



time: float = 0.0
C: float = 1.00000000000000
k1: float = -5.00000000000000
S1: float = 0.0

# Initial assignments
dk1 = 1
J0 = k1
y0 = [S1, k1]
variable_names = ["S1", "k1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, k1 = variables
    dk1: float = 1
    J0: float = k1
    dk1dt: float = dk1
    dS1dt: float = J0
    return dS1dt, dk1dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, k1 = variables
    dk1: float = 1
    J0: float = k1
    return {
        "dk1": dk1,
        "J0": J0,
    }



import math

time: float = 0.0

# Initial assignments
P1 = time
P2 = abs(time)
P3 = abs(time)
P4 = math.acos(time)
P5 = math.acos(-time)
P6 = math.asin(time)
P7 = -math.asin(time)
P8 = math.atan(time)
P9 = -math.atan(time)
P10 = math.ceil(time)
P11 = math.ceil(-time)
P13 = math.cos(time)
P14 = math.cos(time)
P15 = math.exp(time)
P16 = math.exp(-time)
P18 = math.floor(time)
P19 = math.floor(-time)
P20 = math.log(time + 1)
P22 = math.log(time + 1) / math.log(10)
P24 = time**2
P25 = 2**time
P26 = time**time
P29 = math.sqrt(time)
P31 = math.sin(time)
P32 = -math.sin(time)
P34 = math.tan(time)
P35 = -math.tan(time)
P37 = time + 2
P38 = time - 2
P39 = (1 / 2) * time
P40 = 3 * time
P41 = time + 2
P42 = 2 - time
P43 = 2 / (time + 1)
P44 = 3 * time
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    P1: float = time
    P2: float = abs(time)
    P3: float = abs(time)
    P4: float = math.acos(time)
    P5: float = math.acos(-time)
    P6: float = math.asin(time)
    P7: float = -math.asin(time)
    P8: float = math.atan(time)
    P9: float = -math.atan(time)
    P10: float = math.ceil(time)
    P11: float = math.ceil(-time)
    P13: float = math.cos(time)
    P14: float = math.cos(time)
    P15: float = math.exp(time)
    P16: float = math.exp(-time)
    P18: float = math.floor(time)
    P19: float = math.floor(-time)
    P20: float = math.log(time + 1)
    P22: float = math.log(time + 1) / math.log(10)
    P24: float = time**2
    P25: float = 2**time
    P26: float = time**time
    P29: float = math.sqrt(time)
    P31: float = math.sin(time)
    P32: float = -math.sin(time)
    P34: float = math.tan(time)
    P35: float = -math.tan(time)
    P37: float = time + 2
    P38: float = time - 2
    P39: float = (1 / 2) * time
    P40: float = 3 * time
    P41: float = time + 2
    P42: float = 2 - time
    P43: float = 2 / (time + 1)
    P44: float = 3 * time
    return {
        "P1": P1,
        "P2": P2,
        "P3": P3,
        "P4": P4,
        "P5": P5,
        "P6": P6,
        "P7": P7,
        "P8": P8,
        "P9": P9,
        "P10": P10,
        "P11": P11,
        "P13": P13,
        "P14": P14,
        "P15": P15,
        "P16": P16,
        "P18": P18,
        "P19": P19,
        "P20": P20,
        "P22": P22,
        "P24": P24,
        "P25": P25,
        "P26": P26,
        "P29": P29,
        "P31": P31,
        "P32": P32,
        "P34": P34,
        "P35": P35,
        "P37": P37,
        "P38": P38,
        "P39": P39,
        "P40": P40,
        "P41": P41,
        "P42": P42,
        "P43": P43,
        "P44": P44,
    }



time: float = 0.0
k1: float = 75.0000000000000
k2: float = 25.0000000000000
S1: float = 0.0100000000000000
S2: float = 0.0100000000000000
S3: float = 0.0200000000000000
S4: float = 0.0100000000000000

# Initial assignments
dS1 = -S1 * S2 * k1 + S3 * S4 * k2
dS2 = -S1 * S2 * k1 + S3 * S4 * k2
dS3 = S1 * S2 * k1 - S3 * S4 * k2
dS4 = S1 * S2 * k1 - S3 * S4 * k2
y0 = [S1, S2, S3, S4]
variable_names = ["S1", "S2", "S3", "S4"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3, S4 = variables
    dS1: float = -S1 * S2 * k1 + S3 * S4 * k2
    dS2: float = -S1 * S2 * k1 + S3 * S4 * k2
    dS3: float = S1 * S2 * k1 - S3 * S4 * k2
    dS4: float = S1 * S2 * k1 - S3 * S4 * k2
    dS1dt: float = dS1
    dS2dt: float = dS2
    dS3dt: float = dS3
    dS4dt: float = dS4
    return dS1dt, dS2dt, dS3dt, dS4dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3, S4 = variables
    dS1: float = -S1 * S2 * k1 + S3 * S4 * k2
    dS2: float = -S1 * S2 * k1 + S3 * S4 * k2
    dS3: float = S1 * S2 * k1 - S3 * S4 * k2
    dS4: float = S1 * S2 * k1 - S3 * S4 * k2
    return {
        "dS1": dS1,
        "dS2": dS2,
        "dS3": dS3,
        "dS4": dS4,
    }



time: float = 0.0
k1: float = 0.500000000000000
k2: float = 50.0000000000000

# Initial assignments
C = (1 / 50) * k2
S1_amount = 1.0 * C
S2_amount = 1.5 * C
S1 = S1_amount / C
S2 = S2_amount / C
reaction1 = C * S1 * k1
y0 = [S1_amount, S2_amount]
variable_names = ["S1_amount", "S2_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_amount, S2_amount = variables
    S1: float = S1_amount / C
    reaction1: float = C * S1 * k1
    dS1_amountdt: float = -reaction1
    dS2_amountdt: float = reaction1
    return dS1_amountdt, dS2_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_amount, S2_amount = variables
    S1: float = S1_amount / C
    S2: float = S2_amount / C
    reaction1: float = C * S1 * k1
    return {
        "S1": S1,
        "S2": S2,
        "reaction1": reaction1,
    }



time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
p1: float = 0.250000000000000

# Initial assignments
C = 2 * p1
S1_amount = 0.001 * C
S2_amount = 0.002 * C
S3_amount = 0.001 * C
S1 = S1_amount / C
S2 = S2_amount / C
S3 = S3_amount / C
reaction1 = C * S1 * S2 * k1
reaction2 = C * S3 * k2
y0 = [S1_amount, S2_amount, S3_amount]
variable_names = ["S1_amount", "S2_amount", "S3_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_amount, S2_amount, S3_amount = variables
    S1: float = S1_amount / C
    S2: float = S2_amount / C
    S3: float = S3_amount / C
    reaction1: float = C * S1 * S2 * k1
    reaction2: float = C * S3 * k2
    dS1_amountdt: float = -reaction1 + reaction2
    dS2_amountdt: float = -reaction1 + reaction2
    dS3_amountdt: float = reaction1 - reaction2
    return dS1_amountdt, dS2_amountdt, dS3_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_amount, S2_amount, S3_amount = variables
    S1: float = S1_amount / C
    S2: float = S2_amount / C
    S3: float = S3_amount / C
    reaction1: float = C * S1 * S2 * k1
    reaction2: float = C * S3 * k2
    return {
        "S1": S1,
        "S2": S2,
        "S3": S3,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
C: float = 1.00000000000000
S1_create: float = 1.00000000000000
S1: float = 2.00000000000000

# Initial assignments
S1_conc = S1 / C
dS1_create = 1
J0 = 1
J1 = 1
S1_degrade = 3
y0 = [S1, S1_create]
variable_names = ["S1", "S1_create"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S1_create = variables
    dS1_create: float = 1
    J0: float = 1
    J1: float = 1
    dS1_createdt: float = dS1_create
    dS1dt: float = -C * J0 * S1_degrade + C * J1 * S1_create
    return dS1dt, dS1_createdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S1_create = variables
    S1_conc: float = S1 / C
    dS1_create: float = 1
    J0: float = 1
    J1: float = 1
    return {
        "S1_conc": S1_conc,
        "dS1_create": dS1_create,
        "J0": J0,
        "J1": J1,
    }



time: float = 0.0
C: float = 2.00000000000000
J1_S1_stoich: float = 0.0100000000000000
S1_stoich: float = 2.00000000000000

# Initial assignments
J0 = 0.0100000000000000
J1 = J1_S1_stoich
k0 = 2 * S1_stoich
S1_amount = 2.0 * C
S2_amount = 3.0 * C
S1 = S1_amount / C
S2 = S2_amount / C
y0 = [S1_amount, S2_amount]
variable_names = ["S1_amount", "S2_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_amount, S2_amount = variables
    J0: float = 0.0100000000000000
    J1: float = J1_S1_stoich
    dS1_amountdt: float = -C * J0 * S1_stoich
    dS2_amountdt: float = C * J1
    return dS1_amountdt, dS2_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_amount, S2_amount = variables
    S1: float = S1_amount / C
    S2: float = S2_amount / C
    J0: float = 0.0100000000000000
    J1: float = J1_S1_stoich
    return {
        "S1": S1,
        "S2": S2,
        "J0": J0,
        "J1": J1,
    }



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import wadler_lindig as wl

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from pysbml.parse.mathml import Base, Symbol


__all__ = [
    "Assignment",
    "AtomicUnit",
    "Compartment",
    "CompositeUnit",
    "Constraint",
    "Delay",
    "Derived",
    "Document",
    "Event",
    "Function",
    "Model",
    "Parameter",
    "Plugin",
    "Priority",
    "Reaction",
    "Species",
    "Trigger",
    "md_table_from_dict",
]


def md_table_from_dict(
    headers: list[str],
    els: Iterable[Iterable[Any]],
) -> str:
    cols = len(headers)
    fmt = "| {} | ".format(" | ".join("{}" for _ in range(cols)))

    top = fmt.format(*headers)
    div = fmt.format(*("---" for _ in range(cols)))
    return "\n".join((top, div, *(fmt.format(*k) for k in els)))


@dataclass
class Plugin:
    name: str

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass
class Document:
    model: Model
    plugins: list[Plugin]

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class AtomicUnit:
    kind: str
    exponent: int
    scale: int
    multiplier: float

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class CompositeUnit:
    sbml_id: str
    units: list

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Parameter:
    value: float
    is_constant: bool
    unit: str | None

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Compartment:
    name: str
    dimensions: int
    size: float
    units: str
    is_constant: bool

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Species:
    compartment: str | None
    conversion_factor: str | None
    initial_amount: float | None
    initial_concentration: float | None
    substance_units: str | None
    has_only_substance_units: bool
    has_boundary_condition: bool
    is_constant: bool

    def is_concentration(self) -> bool:
        if self.initial_concentration is not None:
            return True
        if self.has_only_substance_units:
            return False
        return False

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Derived:
    body: Base
    args: list[Symbol]

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Function:
    body: Base
    args: list[Symbol]

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass(kw_only=True, slots=True)
class Reaction:
    body: Base
    stoichiometry: Mapping[str, float | list[tuple[float, str]]]
    args: list[Symbol]
    local_pars: dict[str, Parameter] = field(default_factory=dict)

    def __repr__(self) -> str:
        return wl.pformat(self)


@dataclass
class Trigger:
    math: Base | None
    args: list[Symbol]
    initial_value: bool
    persistent: bool


@dataclass
class Delay:
    math: Base | None
    args: list[Symbol]


@dataclass
class Priority:
    math: Base | None
    args: list[Symbol]


@dataclass
class Assignment:
    variable: str
    math: Base | None
    args: list[Symbol]


@dataclass
class Event:
    assignments: list[Assignment]
    trigger: Trigger | None
    delay: Delay | None
    priority: Priority | None


@dataclass
class Constraint:
    math: Base | None
    args: list[Symbol]
    message: str


@dataclass(kw_only=True, slots=True)
class Model:
    name: str
    conversion_factor: str | None = None
    # Collections
    boundary_species: set[str] = field(default_factory=set)
    # Parsed stuff
    events: dict[str, Event] = field(default_factory=dict)
    constraints: dict[str, Constraint] = field(default_factory=dict)
    atomic_units: dict[str, AtomicUnit] = field(default_factory=dict)
    composite_units: dict[str, CompositeUnit] = field(default_factory=dict)
    compartments: dict[str, Compartment] = field(default_factory=dict)
    parameters: dict[str, Parameter] = field(default_factory=dict)
    variables: dict[str, Species] = field(default_factory=dict)
    assignment_rules: dict[str, Derived] = field(default_factory=dict)
    algebraic_rules: dict[str, Derived] = field(default_factory=dict)
    rate_rules: dict[str, Derived] = field(default_factory=dict)
    initial_assignments: dict[str, Derived] = field(default_factory=dict)
    functions: dict[str, Function] = field(default_factory=dict)
    reactions: dict[str, Reaction] = field(default_factory=dict)

    def __repr__(self) -> str:
        return wl.pformat(self)

    def _repr_markdown_(self) -> str:
        content = [f"# {self.name}"]
        if len(self.functions) > 0:
            content.append("# Functions")
            content.append(
                md_table_from_dict(
                    headers=["name", "args", "body"],
                    els=[(k, v.args, v.body) for k, v in self.functions.items()],
                )
            )
        if len(self.compartments) > 0:
            content.append("# Compartment")
            content.append(
                md_table_from_dict(
                    headers=["name", "size", "is_constant"],
                    els=[
                        (k, v.size, v.is_constant) for k, v in self.compartments.items()
                    ],
                )
            )
        if len(self.variables) > 0:
            content.append("# Variables")
            content.append(
                md_table_from_dict(
                    headers=[
                        "name",
                        "amount",
                        "conc",
                        "constant",
                        "substance_units",
                        "compartment",
                        "only_substance_units",
                        "boundary_condition",
                    ],
                    els=[
                        (
                            k,
                            v.initial_amount,
                            v.initial_concentration,
                            v.is_constant,
                            v.substance_units,
                            v.compartment,
                            v.has_only_substance_units,
                            v.has_boundary_condition,
                        )
                        for k, v in self.variables.items()
                    ],
                )
            )
        if len(self.parameters) > 0:
            content.append("# Parameters")
            content.append(
                md_table_from_dict(
                    headers=["name", "value", "is_constant", "unit"],
                    els=[
                        (k, v.value, v.is_constant, v.unit)
                        for k, v in self.parameters.items()
                    ],
                )
            )
        if len(self.assignment_rules) > 0:
            content.append("# Assignment rules")
            content.append(
                md_table_from_dict(
                    headers=["name", "args", "body"],
                    els=[(k, v.args, v.body) for k, v in self.assignment_rules.items()],
                )
            )
        if len(self.algebraic_rules) > 0:
            content.append("# Algebraic rules")
            content.append(
                md_table_from_dict(
                    headers=["name", "args", "body"],
                    els=[(k, v.args, v.body) for k, v in self.algebraic_rules.items()],
                )
            )
        if len(self.rate_rules) > 0:
            content.append("# Rate_rules")
            content.append(
                md_table_from_dict(
                    headers=["name", "args", "body"],
                    els=[(k, v.args, v.body) for k, v in self.rate_rules.items()],
                )
            )

        if len(self.reactions) > 0:
            content.append("# Reactions")
            content.append(
                md_table_from_dict(
                    headers=["name", "body", "args", "stoichiometry", "local pars"],
                    els=[
                        (k, v.body, v.args, v.stoichiometry, v.local_pars)
                        for k, v in self.reactions.items()
                    ],
                )
            )
        if len(self.initial_assignments) > 0:
            content.append("# Initial assignments")
            content.append(
                md_table_from_dict(
                    headers=["name", "args", "body"],
                    els=[
                        (k, v.args, v.body) for k, v in self.initial_assignments.items()
                    ],
                )
            )
        return "\n".join(content)



time: float = 0.0
c: float = 2.00000000000000

# Initial assignments
dS1 = 1
S1 = 0.6 * c
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    dS1: float = 1
    dS1dt: float = dS1
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    dS1: float = 1
    return {
        "dS1": dS1,
    }



time: float = 0.0

# Initial assignments
p1 = time % 2
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    p1: float = time % 2
    return {
        "p1": p1,
    }



from __future__ import annotations

import itertools as it
import logging
import operator as op
from functools import reduce
from typing import TYPE_CHECKING, Any

import sympy

from pysbml.parse import mathml

__all__ = ["LOGGER", "convert_mathml"]

if TYPE_CHECKING:
    from pysbml.transform import data

LOGGER = logging.getLogger(__name__)


def _handle_node(
    node: mathml.Base,
    fns: dict[str, data.Expr],
    *,
    as_bool: bool = False,
    as_number: bool = False,
) -> Any:
    match node:
        case mathml.Symbol(name):
            return sympy.Symbol(name)
        case mathml.Constant(name):
            if name == "e":
                return sympy.E
            if name == "pi":
                return sympy.pi
            msg = f"Unknown constant {name}"
            raise NotImplementedError(msg)
        case mathml.Boolean(value):
            if as_number:
                return sympy.Float(float(value))
            return sympy.true if value else sympy.false
        case mathml.Integer(value):
            if as_bool:
                return sympy.true if value else sympy.false
            return sympy.Integer(value)
        case mathml.Float(value):
            if as_bool:
                return sympy.true if value else sympy.false
            return sympy.Float(value)
        # unary
        case mathml.Abs(value):
            return sympy.Abs(_handle_node(value, fns, as_number=True))
        case mathml.Ceiling(value):
            return sympy.ceiling(_handle_node(value, fns, as_number=True))
        case mathml.Exp(value):
            return sympy.exp(_handle_node(value, fns, as_number=True))
        case mathml.Factorial(value):
            return sympy.factorial(_handle_node(value, fns, as_number=True))
        case mathml.Floor(value):
            return sympy.floor(_handle_node(value, fns, as_number=True))
        case mathml.Ln(value):
            return sympy.ln(_handle_node(value, fns, as_number=True))
        case mathml.Log(base, value):
            return sympy.log(
                _handle_node(value, fns, as_number=True),
                _handle_node(base, fns, as_number=True),
            )
        case mathml.Sqrt(base, value):
            return sympy.root(
                _handle_node(value, fns, as_number=True),
                _handle_node(base, fns, as_number=True),
            )
        case mathml.Sin(value):
            return sympy.sin(_handle_node(value, fns, as_number=True))
        case mathml.Cos(value):
            return sympy.cos(_handle_node(value, fns, as_number=True))
        case mathml.Tan(value):
            return sympy.tan(_handle_node(value, fns, as_number=True))
        case mathml.Sec(value):
            return sympy.sec(_handle_node(value, fns, as_number=True))
        case mathml.Csc(value):
            return sympy.csc(_handle_node(value, fns, as_number=True))
        case mathml.Cot(value):
            return sympy.cot(_handle_node(value, fns, as_number=True))
        case mathml.Asin(value):
            return sympy.asin(_handle_node(value, fns, as_number=True))
        case mathml.Acos(value):
            return sympy.acos(_handle_node(value, fns, as_number=True))
        case mathml.Atan(value):
            return sympy.atan(_handle_node(value, fns, as_number=True))
        case mathml.Acot(value):
            return sympy.acot(_handle_node(value, fns, as_number=True))
        case mathml.ArcSec(value):
            return sympy.asec(_handle_node(value, fns, as_number=True))
        case mathml.ArcCsc(value):
            return sympy.acsc(_handle_node(value, fns, as_number=True))
        case mathml.Sinh(value):
            return sympy.sinh(_handle_node(value, fns, as_number=True))
        case mathml.Cosh(value):
            return sympy.cosh(_handle_node(value, fns, as_number=True))
        case mathml.Tanh(value):
            return sympy.tanh(_handle_node(value, fns, as_number=True))
        case mathml.Sech(value):
            return sympy.sech(_handle_node(value, fns, as_number=True))
        case mathml.Csch(value):
            return sympy.csch(_handle_node(value, fns, as_number=True))
        case mathml.Coth(value):
            return sympy.coth(_handle_node(value, fns, as_number=True))
        case mathml.ArcSinh(value):
            return sympy.asinh(_handle_node(value, fns, as_number=True))
        case mathml.ArcCosh(value):
            return sympy.acosh(_handle_node(value, fns, as_number=True))
        case mathml.ArcTanh(value):
            return sympy.atanh(_handle_node(value, fns, as_number=True))
        case mathml.ArcCsch(value):
            return sympy.acsch(_handle_node(value, fns, as_number=True))
        case mathml.ArcSech(value):
            return sympy.asech(_handle_node(value, fns, as_number=True))
        case mathml.ArcCoth(value):
            return sympy.acoth(_handle_node(value, fns, as_number=True))
        case mathml.Lambda(fn, args):
            return sympy.Lambda(
                tuple(_handle_node(i, fns) for i in args), _handle_node(fn, fns)
            )
        # binary
        case mathml.Pow(left, right):
            return sympy.Pow(
                _handle_node(left, fns, as_number=True),
                _handle_node(right, fns, as_number=True),
            )
        case mathml.Implies(left, right):
            raise NotImplementedError
        # n-ary
        case mathml.Function(name, children):
            fn = fns[name]
            return fn(*(_handle_node(i, fns) for i in children))  # type: ignore
        case mathml.Max(children):
            return sympy.Max(*(_handle_node(i, fns, as_number=True) for i in children))
        case mathml.Min(children):
            return sympy.Min(*(_handle_node(i, fns, as_number=True) for i in children))
        case mathml.Rem(children):
            return reduce(
                op.mod,
                (_handle_node(i, fns, as_number=True) for i in children),
            )
        case mathml.And(children):
            return sympy.And(*(_handle_node(i, fns, as_bool=True) for i in children))
        case mathml.Not(children):
            return sympy.Not(*(_handle_node(i, fns, as_bool=True) for i in children))
        case mathml.Or(children):
            return sympy.Or(*(_handle_node(i, fns, as_bool=True) for i in children))
        case mathml.Xor(children):
            return sympy.Xor(*(_handle_node(i, fns, as_bool=True) for i in children))
        case mathml.Eq(children):
            return reduce(
                op.and_,
                (
                    sympy.Eq(a, b)
                    for a, b in it.pairwise(_handle_node(i, fns) for i in children)
                ),
            )
        case mathml.NotEqual(children):
            return reduce(
                op.and_,
                (
                    sympy.Unequality(a, b)
                    for a, b in it.pairwise(_handle_node(i, fns) for i in children)
                ),
            )
        case mathml.GreaterEqual(children):
            return reduce(
                op.and_,
                (
                    a >= b
                    for a, b in it.pairwise(_handle_node(i, fns) for i in children)
                ),
            )
        case mathml.GreaterThan(children):
            return reduce(
                op.and_,
                (a > b for a, b in it.pairwise(_handle_node(i, fns) for i in children)),
            )
        case mathml.LessEqual(children):
            return reduce(
                op.and_,
                (
                    a <= b
                    for a, b in it.pairwise(_handle_node(i, fns) for i in children)
                ),
            )
        case mathml.LessThan(children):
            return reduce(
                op.and_,
                (a < b for a, b in it.pairwise(_handle_node(i, fns) for i in children)),
            )
        case mathml.Piecewise(children):
            # Only treat the second arg is necessarily bool
            handled = [
                _handle_node(i, fns)
                if x % 2 == 0
                else _handle_node(i, fns, as_bool=True)
                for x, i in enumerate(children)
            ]
            pairs = [
                (handled[2 * i], handled[2 * i + 1]) for i in range(len(children) // 2)
            ]
            return sympy.Piecewise(*pairs, (handled[-1], True))
        case mathml.Add(children):
            return sympy.Add(*(_handle_node(i, fns, as_number=True) for i in children))
        case mathml.Mul(children):
            return sympy.Mul(*(_handle_node(i, fns, as_number=True) for i in children))
        # These need special handling as sympy doesn't have a nice
        # constructor
        case mathml.Minus(children):
            children = [_handle_node(i, fns, as_number=True) for i in children]
            if len(children) == 1:
                return -children[0]
            return reduce(op.sub, children)
        case mathml.Divide(children):
            return reduce(
                op.truediv, (_handle_node(i, fns, as_number=True) for i in children)
            )
        case mathml.IntDivide(children):
            return reduce(
                op.floordiv, (_handle_node(i, fns, as_number=True) for i in children)
            )
        case _:
            raise NotImplementedError(type(node))


def convert_mathml(node: mathml.Base, fns: dict[str, data.Expr]) -> sympy.Expr:
    return _handle_node(node=node, fns=fns)



time: float = 0.0

# Initial assignments
f = 2
e = f + 1
d = e + 1
c = d + 1
b = c + 1
a = b + 1
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    b: float = c + 1
    d: float = e + 1
    f: float = 2
    return {
        "b": b,
        "d": d,
        "f": f,
    }



time: float = 0.0
S1: float = 3.00000000000000
C1: float = 0.500000000000000
S1: float = 3.00000000000000

# Initial assignments
x = S1 / C1
S1_conc = S1 / C1
dS1 = 0
dC1 = 0.200000000000000
y0 = [C1, S1]
variable_names = ["C1", "S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    C1, S1 = variables
    dS1: float = 0
    dC1: float = 0.200000000000000
    dS1dt: float = dS1
    dC1dt: float = dC1
    return dC1dt, dS1dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    C1, S1 = variables
    x: float = S1 / C1
    S1_conc: float = S1 / C1
    dS1: float = 0
    dC1: float = 0.200000000000000
    return {
        "x": x,
        "S1_conc": S1_conc,
        "dS1": dS1,
        "dC1": dC1,
    }



time: float = 0.0
k1: float = 0.500000000000000
P1: float = 0.00150000000000000
P2: float = 0.0
P3: float = 0.00100000000000000

# Initial assignments
dP3 = k1 * time
dP1 = -P1 * P3
dP2 = P1 * P3
y0 = [P1, P2, P3]
variable_names = ["P1", "P2", "P3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    P1, P2, P3 = variables
    dP3: float = k1 * time
    dP1: float = -P1 * P3
    dP2: float = P1 * P3
    dP3dt: float = dP3
    dP1dt: float = dP1
    dP2dt: float = dP2
    return dP1dt, dP2dt, dP3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    P1, P2, P3 = variables
    dP3: float = k1 * time
    dP1: float = -P1 * P3
    dP2: float = P1 * P3
    return {
        "dP3": dP3,
        "dP1": dP1,
        "dP2": dP2,
    }



from __future__ import annotations

from sympy import S
from sympy.physics.units import (
    ampere,
    avogadro_number,
    becquerel,
    candela,
    coulomb,
    farad,
    gram,
    gray,
    henry,
    hertz,
    joule,
    katal,
    kelvin,
    kilogram,
    liter,
    lux,
    meter,
    mol,
    newton,
    ohm,
    pascal,
    radian,
    second,
    siemens,
    steradian,
    tesla,
    volt,
    watt,
    weber,
)
from sympy.physics.units.definitions.dimension_definitions import temperature
from sympy.physics.units.prefixes import (
    atto,
    centi,
    deca,
    deci,
    exa,
    femto,
    giga,
    hecto,
    kilo,
    mega,
    micro,
    milli,
    nano,
    peta,
    pico,
    tera,
    yocto,
    yotta,
    zepto,
    zetta,
)
from sympy.physics.units.quantities import Quantity

__all__ = ["CONVERSION", "PREFIXES"]

celsius = Quantity("celsius", abbrev="°C")
celsius.set_global_dimension(temperature)
# we can't do the proper offset here, sympy only allows proportional
# changes
celsius.set_global_relative_scale_factor(S.One, kelvin)

CONVERSION: dict[str, Quantity] = {
    "AMPERE": ampere,
    "AVOGADRO": avogadro_number,
    "BECQUEREL": becquerel,
    "CANDELA": candela,
    "CELSIUS": celsius,
    "COULOMB": coulomb,
    # "DIMENSIONLESS": ,
    "FARAD": farad,
    "GRAM": gram,
    "GRAY": gray,
    "HENRY": henry,
    "HERTZ": hertz,
    # "ITEM": ,
    "JOULE": joule,
    "KATAL": katal,
    "KELVIN": kelvin,
    "KILOGRAM": kilogram,
    "LITER": liter,
    "LITRE": liter,
    # "LUMEN": lumen,
    "LUX": lux,
    "METER": meter,
    "METRE": meter,
    "MOLE": mol,
    "NEWTON": newton,
    "OHM": ohm,
    "PASCAL": pascal,
    "RADIAN": radian,
    "SECOND": second,
    "SIEMENS": siemens,
    # "SIEVERT": sievert,
    "STERADIAN": steradian,
    "TESLA": tesla,
    "VOLT": volt,
    "WATT": watt,
    "WEBER": weber,
    # "INVALID": ,
}

PREFIXES = {
    24: yotta,
    21: zetta,
    18: exa,
    15: peta,
    12: tera,
    9: giga,
    6: mega,
    3: kilo,
    2: hecto,
    1: deca,
    -1: deci,
    -2: centi,
    -3: milli,
    -6: micro,
    -9: nano,
    -12: pico,
    -15: femto,
    -18: atto,
    -21: zepto,
    -24: yocto,
}



from __future__ import annotations

import libsbml

__all__ = ["get_ast_types", "get_operator_mappings", "get_unit_conversion"]

_unit_conversion = {
    "UNIT_KIND_AMPERE": "AMPERE",
    "UNIT_KIND_AVOGADRO": "AVOGADRO",
    "UNIT_KIND_BECQUEREL": "BECQUEREL",
    "UNIT_KIND_CANDELA": "CANDELA",
    "UNIT_KIND_CELSIUS": "CELSIUS",
    "UNIT_KIND_COULOMB": "COULOMB",
    "UNIT_KIND_DIMENSIONLESS": "DIMENSIONLESS",
    "UNIT_KIND_FARAD": "FARAD",
    "UNIT_KIND_GRAM": "GRAM",
    "UNIT_KIND_GRAY": "GRAY",
    "UNIT_KIND_HENRY": "HENRY",
    "UNIT_KIND_HERTZ": "HERTZ",
    "UNIT_KIND_ITEM": "ITEM",
    "UNIT_KIND_JOULE": "JOULE",
    "UNIT_KIND_KATAL": "KATAL",
    "UNIT_KIND_KELVIN": "KELVIN",
    "UNIT_KIND_KILOGRAM": "KILOGRAM",
    "UNIT_KIND_LITER": "LITER",
    "UNIT_KIND_LITRE": "LITER",
    "UNIT_KIND_LUMEN": "LUMEN",
    "UNIT_KIND_LUX": "LUX",
    "UNIT_KIND_METER": "METER",
    "UNIT_KIND_METRE": "METRE",
    "UNIT_KIND_MOLE": "MOLE",
    "UNIT_KIND_NEWTON": "NEWTON",
    "UNIT_KIND_OHM": "OHM",
    "UNIT_KIND_PASCAL": "PASCAL",
    "UNIT_KIND_RADIAN": "RADIAN",
    "UNIT_KIND_SECOND": "SECOND",
    "UNIT_KIND_SIEMENS": "SIEMENS",
    "UNIT_KIND_SIEVERT": "SIEVERT",
    "UNIT_KIND_STERADIAN": "STERADIAN",
    "UNIT_KIND_TESLA": "TESLA",
    "UNIT_KIND_VOLT": "VOLT",
    "UNIT_KIND_WATT": "WATT",
    "UNIT_KIND_WEBER": "WEBER",
    "UNIT_KIND_INVALID": "INVALID",
}


def get_unit_conversion() -> dict[str, str]:
    return {getattr(libsbml, k): v for k, v in _unit_conversion.items()}


def get_ast_types() -> dict[str, str]:
    return dict(
        sorted(
            {
                getattr(libsbml, i): i for i in dir(libsbml) if i.startswith("AST_")
            }.items()
        )
    )


def get_operator_mappings() -> dict[str, str]:
    return {
        "AST_TIMES": "*",
        "AST_PLUS": "+",
        "AST_MINUS": "-",
        "AST_DIVIDE": "/",
        "AST_POWER": "**",
    }



time: float = 0.0
k1: float = 0.350000000000000
k2: float = 180.000000000000
compartment: float = 0.300000000000000
S1: float = 0.000150000000000000
S2: float = 0.0

# Initial assignments
S1_conc = S1 / compartment
S2_conc = S2 / compartment
reaction1 = S1_conc * k1
reaction2 = S2_conc**2 * k2
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    reaction1: float = S1_conc * k1
    reaction2: float = S2_conc**2 * k2
    dS1dt: float = -compartment * reaction1 + compartment * reaction2
    dS2dt: float = 2.0 * compartment * reaction1 - 2.0 * compartment * reaction2
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    reaction1: float = S1_conc * k1
    reaction2: float = S2_conc**2 * k2
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
C: float = 1.00000000000000
S1_stoich: float = 1.00000000000000
S1: float = 2.00000000000000

# Initial assignments
S1_conc = S1 / C
dS1_stoich = 1.50000000000000
J0 = 0.100000000000000
y0 = [S1, S1_stoich]
variable_names = ["S1", "S1_stoich"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S1_stoich = variables
    dS1_stoich: float = 1.50000000000000
    J0: float = 0.100000000000000
    dS1_stoichdt: float = dS1_stoich
    dS1dt: float = C * J0 * S1_stoich
    return dS1dt, dS1_stoichdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S1_stoich = variables
    S1_conc: float = S1 / C
    dS1_stoich: float = 1.50000000000000
    J0: float = 0.100000000000000
    return {
        "S1_conc": S1_conc,
        "dS1_stoich": dS1_stoich,
        "J0": J0,
    }



time: float = 0.0
C: float = 1.00000000000000
S2_create: float = 1.00000000000000
S1: float = 30.0000000000000
S2: float = 0.0

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
dS2_create = 1
J0 = 1
S1_degrade = 3
y0 = [S1, S2, S2_create]
variable_names = ["S1", "S2", "S2_create"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S2_create = variables
    dS2_create: float = 1
    J0: float = 1
    dS2_createdt: float = dS2_create
    dS1dt: float = -C * J0 * S1_degrade
    dS2dt: float = C * J0 * S2_create
    return dS1dt, dS2dt, dS2_createdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S2_create = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    dS2_create: float = 1
    J0: float = 1
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "dS2_create": dS2_create,
        "J0": J0,
    }



import math

time: float = 0.0
k1: float = 1.54600000000000
compartment: float = math.nan
S3: float = 0.0
S1: float = 1.50000000000000
S2: float = 0.0

# Initial assignments
dS3 = 0.650000000000000
reaction1 = S1 * k1
y0 = [S1, S2, S3]
variable_names = ["S1", "S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3 = variables
    dS3: float = 0.650000000000000
    reaction1: float = S1 * k1
    dS3dt: float = dS3
    dS1dt: float = -reaction1
    dS2dt: float = reaction1
    return dS1dt, dS2dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3 = variables
    dS3: float = 0.650000000000000
    reaction1: float = S1 * k1
    return {
        "dS3": dS3,
        "reaction1": reaction1,
    }
